//
//                     TxWin, Textmode Windowing Library
//
//   Original code Copyright (c) 1995-2021 Fsys Software and Jan van Wijk
//
// ==========================================================================
//
//   TxLib, released under MIT License
//
//   Copyright (c) 1995-2021  Fsys Software and Jan Van Wijk
//
//   Permission is hereby granted, free of charge, to any person obtaining a copy
//   of this software and associated documentation files (the "Software"), to deal
//   in the Software without restriction, including without limitation the rights
//   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//   copies of the Software, and to permit persons to whom the Software is
//   furnished to do so, subject to the following conditions:
//
//   The above copyright notice and this permission notice shall be included in all
//   copies or substantial portions of the Software.
//
//   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
//   SOFTWARE.
//
//
//   Questions on TxWin licensing can be directed to: info@dfsee.com
//
// ==========================================================================
//
// TxLib command, argv/argc and number parsing
//
// JvW  28-02-2021 LICENSING: Changed from LGPL to the more liberal MIT license
// JvW  04-07-2007 Interpret -x123 as longname, not -x:123   (DFSee 9.00)
// JvW  21-11-2004 Allow explicit decimal prefix 0n and 0t   (DFSee 6.16)
// JvW  08-08-2003 Long-option support and allow '/' switch  (DFSee 5.25)
// JvW  15-12-2001 No string-value for -tValue type syntax   (DFSee 5.08)
// JvW  04-12-2001 Make numeric 0 the default for option -o  (DFSee 4.11)
// JvW  17-11-2001 Added '' as alternative string syntax     (DFSee 4.10)
// JvW  05-11-2001 Added -tValue and -t- type of syntax      (DFSee 4.09)
// JvW  27-08-2001 Added argv/argc as input alternative      (DFSee 4.03)
// JvW  13-08-2001 First released version                    (DFSee 4.02)

#include <txlib.h>                              // TxLib interface
#include <txwpriv.h>                            // TXW private interfaces
#include <txtpriv.h>                            // TXT private interfaces

#define TXA_SIGNATURE   0xFACEBEEF              // valid magic signature value

       TXA_ELEMENT  *txacur = NULL;             // current, top of stack
       TXA_ELEMENT  *txa1st = NULL;             // first, bottom of stack

typedef char TXA_LONGOPT[ TXA_O_LEN +1];

static TXA_LONGOPT   txalongopt[TXA_LSIZE] =
{                                               // predefined long names
   "all",
   "auto",
   "clear",
   "color",
   "config",
   "debug",
   "dir",
   "disk",
   "entry",
   "file",
   "first",
   "fix",
   "ini",
   "keyb",
   "label",
   "last",
   "lines",
   "list",
   "manual",
   "menu",
   "mouse",
   "name",
   "new",
   "query",
   "refresh",
   "reset",
   "scheme",
   "screen",
   "set",
   "simulate",
   "style",
   "test",
   "type",
   "trace",
   "update",
   "url",
   "xml",
   "raw",
   "echo",
   "prompt",
   "sigint",
   "insert",
   "size",
   "sysmouse",
   "node",
   "header",
   "silent",
   "",                                          // last TX reserved one

   "",                                          // TXA_O_APP0
   "",
   "",
   "",
   "",
   "",
   "",
   "",
   "",
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""
   ""                                           // TXA_O_AP31
};



/*****************************************************************************/
// Set or query value for a specific LONGNAME value (the LONG option name)
/*****************************************************************************/
char  *TxaOptionLongName                        // RET   resulting option name
(
   char                opt,                     // IN    TXA_O_ option value
   char               *name                     // IN    long name for option
)                                               //       or NULL to query
{
   char               *rc = NULL;               // function return
   int                 lni;

   ENTER();

   if ((opt >= TXA_LBASE) && ((lni = (opt - TXA_LBASE)) < TXA_LSIZE))
   {
      if (name != NULL)
      {
         TxCopy( txalongopt[lni], name, TXA_O_LEN +1);
      }
      rc = &(txalongopt[lni][0]);
      TRACES(("Name for APP option %u set to '%s'\n", opt - TXA_O_APP0, rc));
   }
   RETURN (rc);
}                                               // end 'TxaOptionLongName'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Create new level of command options for a set of argc/argv variables
/*****************************************************************************/
ULONG TxaParseArgcArgv                          // RET   result
(
   int                 argc,                    // IN    argument count
   char               *argv[],                  // IN    array of arguments
   char               *exename,                 // IN    alternative argv[0]
   BOOL                freeform,                // IN    free format options
   TXHANDLE           *txh                      // OUT   TXA handle (optional)
)
{
   ULONG               rc;
   TXA_ELEMENT        *txa;
   char               *item;                    // parsed item
   int                 len;                     // length of item
   int                 i;

   ENTER();

   if ((rc = TxaNewParseElement( &txa)) == NO_ERROR)
   {
      for (i = 0; (i < argc) && (rc == NO_ERROR); i++)
      {
         if ((i == 0) && (exename != NULL))
         {
            item = exename;                     // preserve alternative argv[0]
         }
         else
         {
            item = argv[i];
         }
         len  = strlen( item);

         rc = txaReadAndStoreItem( item, len, freeform, TRUE, txa);
      }
   }
   if (txh != NULL)
   {
      *txh = (TXHANDLE) txa;
   }
   RETURN( rc);
}                                               // end 'TxaParseArgcArgv'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Parse a decimal or hex number value from a string
/*****************************************************************************/
ULN64 TxaParseNumber                            // RET   number value
(
   char               *value,                   // IN    value string with nr
   ULONG               rclass,                  // IN    HEX/DEC radix class
   BYTE               *unit                     // OUT   optional unit char, if
)                                               //       not needed use NULL
{
   ULN64               rc = 0;                  // function return
   BOOL                prefix     = TRUE;       // number parsed using prefix
   int                 distance   = 0;          // distance to specific char
   int                 radixclass = rclass;     // class to be used for default

   TRLEVX(700,( "TxaParseNumber: value: '%s'  rclass: 0x%X\n", value, rclass));

   if (unit != NULL)                            // read unit char suffix
   {
      if ((( distance = (int) strcspn( value, ", ")) > 0) &&
          ( value[ distance] == ','))           // unit specifier present
      {
         *unit = value[ distance +1];           // unit character
         if (strchr( "sS", *unit) != NULL)      // 's' unit
         {
            radixclass = TX_RADIX_UN_S_CLASS;   // use sector-unit class
         }
         else
         {
            radixclass = TX_RADIX_UNIT_CLASS;   // use 'other' unit class
         }
      }
      else
      {
         *unit = TXA_DFUNIT;                    // assume default unit
      }
      TRLEVX(700,( "TxaParseNumber: Unit for value: '%s' is: '%c'\n", value, *unit));
   }
   if (*value == '0')                           // possible hex/oct/dec prefix
   {
      switch (tolower(*(value +1)))
      {
         case 'x':
         case 'X': sscanf( value +2, "%llx", &rc); break; // hexadecimal number
         case 'o':
         case 'O': sscanf( value +2, "%llo", &rc); break; // octal number base
         case 't':
         case 'T': sscanf( value +2, "%llu", &rc); break; // decimal number base
         default:  prefix = FALSE;                 break; // to be determined
      }
   }
   else
   {
      prefix = FALSE;
   }
   if (prefix == FALSE)                         // no prefix, test HEX or DEC
   {
      TRLEVX(700,( "value has no prefix ...\n"));
      if ( txIsValidHex( value) &&              // valid HEX upto separator
          !txIsValidDec( value) )               // and has non-decimal digits
      {
         TRLEVX(700,( "TxaParseNumber: auto-interpret as HEX!\n"));
         sscanf( value, "%llx", &rc);           // interpret as HEX
      }
      else                                      // could be both ...
      {
         if (txwa->radixclass & radixclass)     // class set to HEX default
         {
            TRLEVX(700,( "TxaParseNumber: Radix-selection HEX!\n"));
            sscanf( value, "%llx", &rc);        // interpret as HEX
         }
         else
         {
            TRLEVX(700,( "TxaParseNumber: Radix-selection DEC!\n"));
            sscanf( value, "%llu", &rc);        // interpret as DECimal
         }
      }
   }
   else
   {
      TRLEVX(700,( "TxaParseNumber: Selection by prefix '%2.2s'\n", value));
   }
   return (rc);
}                                               // end 'TxaParseNumber'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Parse a 'C' time_t date/time from a string formatted as: EMPTY: 30 days ago
// Nr N: N days ago; +/-N: hours from NOW; else: [yyyy-mm-dd[-| |:]]HH:MM[:SS]
/*****************************************************************************/
time_t TxaParseDateTime                         // RET   date/time value
(
   char               *value                    // IN    value string with nr
)
{
   time_t              rc = 0;                  // default 0 (ALL files, '*')
   time_t              now;
   struct tm          *tstamp;

   ENTER();
   TRACES(( "value: '%s'\n", value));

   time( &now);
   tstamp = localtime( &now);                   // NOW + offset in TM format

   if ((*value == '-') || (*value == '+'))      // hourly offset from now
   {
      rc = now + (atoi( value) * 3600);         // NOW + offset in seconds
   }
   else                                         // (partly) specified date/time
   {
      int count = 0, year = 0, month = 1, day = 1, hour = 0, minute = 0, seconds = 0;

      if (strchr( value, ':') != NULL)          // there is a TIME value
      {
         if (strchr( value, '-') != NULL)       // there is a DATE value
         {
            count = sscanf( value, "%d-%d-%d%*[- :]%d:%d:%d", &year, &month, &day, &hour, &minute, &seconds);
            TRACES(("DT count:%d year:%d month:%d day:%d hour:%d minute:%d, seconds:%d\n",
                                     count, year, month, day, hour, minute, seconds));
            if (count >= 5)
            {
               if (year > 1900)
               {
                  tstamp->tm_year = year - 1900;
               }
               else                             // assume year '20xx'
               {
                  tstamp->tm_year = year + 100;
               }
               tstamp->tm_mon     = month - 1;
               tstamp->tm_mday    = day;
               tstamp->tm_hour    = hour;
               tstamp->tm_min     = minute;
               tstamp->tm_sec     = seconds;
            }
         }
         else                                   // assume just a TIME value
         {                                      // combine with current DATE
            count = sscanf( value, "%d:%d:%d", &hour, &minute, &seconds);
            TRACES(("TM count:%d hour:%d minute:%d, seconds:%d\n", count, hour, minute, seconds));

            if (count >= 2)
            {
               tstamp->tm_hour    = hour;
               tstamp->tm_min     = minute;
               tstamp->tm_sec     = seconds;
            }
         }
         if ((rc = mktime( tstamp)) == -1)
         {
            TRACES(("Error converting TM timestamp!\n"));
            rc = 0;                             // revert oldest timestamp, jan 1 1970
         }
      }
      else
      {
         if (strchr( value, '-') != NULL)       // assume just a DATE or YEAR or YEAR-MONTH value
         {
            count = sscanf( value, "%d-%d-%d", &year, &month, &day);
            TRACES(("DT count:%d year:%d month:%d day:%d\n", count, year, month, day));

            if (count >= 1)
            {
               if (year > 1900)
               {
                  tstamp->tm_year = year - 1900;
               }
               else                             // assume year '20xx'
               {
                  tstamp->tm_year = year + 100;
               }
               tstamp->tm_mon     = month - 1;
               tstamp->tm_mday    = day;
               tstamp->tm_hour    = 0;
               tstamp->tm_min     = 0;
               tstamp->tm_sec     = 0;
               tstamp->tm_isdst   = 0;          // ignore daylight-saving (off by one :)
            }
            if ((rc = mktime( tstamp)) == -1)
            {
               TRACES(("Error converting TM timestamp!\n"));
               rc = 0;                          // revert oldest timestamp, jan 1 1970
            }
         }
         else if (atoi( value) != 0)               // assume single value, days ago
         {
            rc = now - (atoi( value) * 24 * 3600);
         }
      }
   }
   RETURN (rc);
}                                               // end 'TxaParseDateTime'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Read one item (option or argument) and store it in the txa element
/*****************************************************************************/
ULONG txaReadAndStoreItem                       // RET   result
(
   char               *item,                    // IN    item string
   int                 len,                     // IN    length of item
   BOOL                freeform,                // IN    free format options
   BOOL                passthrough,             // IN    keep parameter quotes
   TXA_ELEMENT        *txa                      // INOUT TXA element
)
{
   ULONG               rc = NO_ERROR;
   int                 arguments = txa->argc;   // current number of arguments

   ENTER();

   TRACES(( "Freeform: %s Passthrough: %s Args: %d  Len:%3.3u, Item: '%*.*s'\n",
            (freeform) ? "YES" : "NO ", (passthrough) ? "YES" : "NO ",
             arguments, len, len, len, item));
   if   ((((txa      == txa1st)  &&             // switch on 1st level '/' only
         (*(item)    == '/'))    ||
         (*(item)    == '-'  ))  &&             // switch or option using  '-'
         (*(item +1) != ' '  )   &&             // allowed at this position ?
         (*(item +1) != '\0' )   &&
         (arguments  > 0     )   &&             // never BEFORE 1st word (cmd)
        ((arguments == 1) || freeform))         // possibly after params too
   {
      int              index;                   // index for the option char
      char            *value;                   // ptr to value part of item
      int              distance;                // distance to specific char
      TXA_LONGOPT      lname = {0};             // empty longname to start with

      index = *(item +1);
      if ((index >= TXA_BASE) && (index < (TXA_BASE + TXA_SIZE)))
      {
         index -= TXA_BASE;                     // make it 0 based
         TRACES(( "Direct option index:%3hu\n", index));

         switch (*(item +2))
         {
            case '"': case '\'':                // quoted string
            case '+': case '-':                 // boolean yes/no
               value = item +2;
               break;

            default:
               if ((distance = (int) strcspn( item, ": ")) < len)
               {
                  value = item + distance +1;   // skip the colon ...
               }
               else
               {
                  if ((len > 2) &&              // non-empty option name
                      (item[len -1] == '-'))    // ending in a '-'
                  {
                     distance = len -1;         // length of the name
                     value = item + distance;   // '-' is the value
                  }
                  else
                  {
                     value = item + len;        // no explicit value given
                  }
               }

               if (distance > 2)                // non-trivial longname
               {
                  int  lni;

                  strncpy( lname, item +1, distance -1);
                  lname[distance -1] = 0;       // terminate the name
                  TRACES(( "longname '%s' for: '%s'\n", lname, item));

                  if (strcasecmp( "help", lname) == 0) // predefined, map to '?'
                  {
                     index = '?' - TXA_BASE;
                  }
                  else
                  {
                     for (lni = 0; lni < TXA_LSIZE; lni++)
                     {
                        //- TRACES(("probe '%s' against index: %d = '%s'\n", lname, lni, txalongopt[lni]));
                        if (strlen(txalongopt[lni]) != 0)
                        {
                           if (strcasecmp( txalongopt[lni], lname) == 0)
                           {
                              index = lni + TXA_LBASE - TXA_BASE;

                              TRACES(("Matched defined longname '%s', index: %hu\n",
                                       txalongopt[lni], index));
                           }
                        }
                     }
                  }
               }
               break;
         }
         TxFreeMem(  txa->opt[index].name);     // free existing name, if any
         if (strlen( lname) != 0)               // new longname available ?
         {
            if ((txa->opt[ index].name = TxAlloc(1, strlen(lname) +1)) != NULL)
            {
               strcpy( txa->opt[ index].name, lname);
            }
         }

         if (txa->opt[ index].type == TXA_STRING) // already has a string!
         {
            TRACES(("Free existing string value: '%s'\n",
                       txa->opt[ index].value.string));
            TxFreeMem( txa->opt[ index].value.string);
         }
         txa->opt[ index].type = TXA_NUMBER;
         txa->opt[ index].unit = TXA_DFUNIT;    // default unit
         switch (*value)
         {
            case TXk_SPACE: case '\0':
               txa->opt[ index].type = TXA_NO_VAL;
               txa->opt[ index].value.number = 0;
               TRACES(("NoValueOpt (%3hu='%c')\n", index + TXA_BASE, index + TXA_BASE));
               break;

            case '+': txa->opt[ index].value.number = 1;      break;
            case '-': txa->opt[ index].value.number = 0;      break;
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
               txa->opt[ index].value.number =
                  TxaParseNumber( value, TX_RADIX_STD_CLASS, &(txa->opt[ index].unit));

               TRACES(("NumericOpt (%3hu='%c'), value: %llu\n",
                        index + TXA_BASE, index + TXA_BASE,
                        txa->opt[ index].value.number));
               break;

            default:
               txa->opt[ index].type = TXA_STRING;
               if ((txa->opt[ index].value.string = TxAlloc(1, len +1)) != NULL)
               {
                  if (passthrough)              // straight copy
                  {
                     strcpy(      txa->opt[ index].value.string, value);
                  }
                  else                          // space delimited stuff!
                  {
                     txaCopyItem( txa->opt[ index].value.string, value, len);
                  }

                  TRACES(("String Opt (%3hu='%c') at %8.8x, value: '%s'\n",
                               index + TXA_BASE, index + TXA_BASE,
                               txa->opt[ index].value.string,
                               txa->opt[ index].value.string));
               }
               else
               {
                  rc = TX_ALLOC_ERROR;
               }
               break;
         }
         txa->optc++;                           // count the option
      }
      else                                      // option char out of range
      {
         rc = TX_BAD_OPTION_CHAR;
      }
   }
   else                                         // argv 0 .. n
   {
      if (arguments < TXA_ARGC)                 // room for another ?
      {
         if ((txa->argv[ arguments] = TxAlloc(1, len +1)) != NULL)
         {
            if (passthrough)                    // straight copy
            {
               strcpy(      txa->argv[ arguments], item);
            }
            else                                // space delimited stuff!
            {
               txaCopyItem( txa->argv[ arguments], item, len);
            }
            TRACES(("Stored argv[%d] at %8.8x, value: '%s'\n", arguments,
                         txa->argv[ arguments], txa->argv[ arguments]));
            arguments++;
         }
         else
         {
            rc = TX_ALLOC_ERROR;
         }
      }
      else
      {
         rc = TX_TOO_MANY_ARGS;
      }
   }
   txa->argc = arguments;
   RETURN( rc);
}                                               // end 'txaReadAndStoreItem'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Create new level of command options for a set of argc/argv variables
/*****************************************************************************/
ULONG TxaNewParseElement                        // RET   result
(
   TXA_ELEMENT       **element                  // OUT   TXA element
)
{
   ULONG               rc  = NO_ERROR;
   TXA_ELEMENT        *txa = NULL;

   if ((txa = TxAlloc(1, sizeof(TXA_ELEMENT))) != NULL)
   {
      txa->signature = TXA_SIGNATURE;
      txa->prev      = txacur;

      if (txacur == NULL)                       // first one ?
      {
         txa1st   = txa;
      }
      txacur      = txa;
   }
   else
   {
      rc = TX_ALLOC_ERROR;
   }
   *element = txa;
   return( rc);
}                                               // end 'TxaNewParseElement'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Terminate use of current or all instances of the TXA abstract-data-type
/*****************************************************************************/
BOOL TxaDropParsedCommand                       // RET   more instances left
(
   BOOL                whole_stack              // IN    drop all, terminate
)
{
   TXA_ELEMENT        *txa;
   int                 i;

   ENTER();

   do
   {
      if ((txa = txacur) != NULL)
      {
         for (i = 0; i < TXA_SIZE; i++)         // free allocated options
         {
            if (txa->opt[i].type == TXA_STRING)
            {
               TxFreeMem( txa->opt[i].value.string);
            }
            TxFreeMem( txa->opt[i].name);
         }
         for (i = 0; i < txa->argc; i++)        // free argv storage
         {
            if (txa->argv[i] != NULL)
            {
               TxFreeMem( txa->argv[i]);
            }
         }
         txacur = txa->prev;
         if (txacur == NULL)                    // empty again ?
         {
            txa1st   = NULL;
         }
         TxFreeMem( txa);
      }
   } while (txacur && whole_stack);

   BRETURN((txacur != NULL));
}                                               // end 'TxaDropParsedCommand'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Drop the command-arguments at the current level, but KEEP level and options
/*****************************************************************************/
BOOL TxaDropArgumentsOnly                       // RET   more instances left
(
   void
)
{
   TXA_ELEMENT        *txa;
   int                 i;

   ENTER();

   if ((txa = txacur) != NULL)
   {
      for (i = 0; i < txa->argc; i++)           // free argv storage
      {
         if (txa->argv[i] != NULL)
         {
            TxFreeMem( txa->argv[i]);
         }
      }
      txa->argc = 0;                            // reset to zero arguments
   }
   BRETURN((txacur != NULL));
}                                               // end 'TxaDropArgumentsOnly'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Get reference to specified option data, or NULL if option not set
/*****************************************************************************/
TXA_OPTION *TxaGetOption                        // RET   option ptr or NULL
(
   TXHANDLE            txh,                     // IN    TXA handle
   char                opt                      // IN    option character
)
{
   TXA_OPTION         *rc = NULL;
   TXA_ELEMENT        *txa;                     // specified TXA instance
   int                 i;
   TXA_OPTION         *o;

   if ((txa = txaHandle2Element( txh)) != NULL)
   {
      if (((BYTE) opt >= TXA_BASE) && ((BYTE) opt < (TXA_BASE + TXA_SIZE)))
      {
         i = opt - TXA_BASE;                    // make it 0 based
         o = &txa->opt[ i];
         if (o->type != TXA_NONE)
         {
            rc  = o;
         }
         TRLEVX(200,("TXH (%s): %x option: %3hu = '%c'  value: %llu '%s'\n",
            (txh == TXA_1ST) ? "EXE switch" : "CMD option", txh, opt, opt,
             (o->type == TXA_STRING) ? strlen(o->value.string) : o->value.number,
             (o->type == TXA_NONE  ) ?       "---NONE---"      :
             (o->type != TXA_STRING) ?       "--NUMBER--"      : o->value.string ));
      }
   }
   return( rc);
}                                               // end 'TxaGetOption'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Get simple YES/NO status for specified option. Not set is "NO"
/*****************************************************************************/
BOOL TxaOptionYes                               // RET   option set to YES
(
   TXHANDLE            txh,                     // IN    TXA handle
   char                opt                      // IN    option character
)
{
   BOOL                rc = FALSE;
   TXA_OPTION         *o;

   if ((o = TxaGetOption( txh, opt)) != NULL)
   {
      switch (o->type)
      {
         case TXA_NUMBER:
            if (o->value.number != 0)
            {
               rc = TRUE;
            }
            break;

         case TXA_STRING:
            if ((strcasecmp( o->value.string, "NO")    != 0) &&
                (strcasecmp( o->value.string, "N")     != 0) &&
                (strcasecmp( o->value.string, "0")     != 0) &&
                (strcasecmp( o->value.string, "OFF")   != 0) &&
                (strcasecmp( o->value.string, "FALSE") != 0)  )
            {
               rc = TRUE;
            }
            break;

         default:                               // no value, but set
            rc = TRUE;
            break;
      }
   }
   return( rc);
}                                               // end 'TxaOptionYes'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Get string value for specified option, use default if no string (and warn!)
/*****************************************************************************/
char *TxaOptionStr                              // RET   option String value
(
   TXHANDLE            txh,                     // IN    TXA handle
   char                opt,                     // IN    option character
   char               *error,                   // IN    error text or NULL
   char               *deflt                    // IN    default value
)
{
   char               *rc = deflt;
   TXA_OPTION         *o;

   if ((o = TxaGetOption( txh, opt)) != NULL)
   {
      if (o->type == TXA_STRING)
      {
         rc = o->value.string;
      }
      else                                      // other option type
      {
         if (error != NULL)                     // give error/warning
         {
            TxPrint( "\n%s %s: %s'-%c'%s has no STRING value, using "
                     "default: '%s'\n",
                     (txh != TXA_1ST) ? "Option" : "Switch",
                      error, CBC, opt, CNN, deflt);
         }
      }
   }
   return( rc);
}                                               // end 'TxaOptionStr'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Get number value for specified option, use default if no number (and warn!)
/*****************************************************************************/
ULN64 TxaOptionNum                              // RET   option Number value
(
   TXHANDLE            txh,                     // IN    TXA handle
   char                opt,                     // IN    option character
   char               *error,                   // IN    error text or NULL
   ULN64               deflt                    // IN    default value
)
{
   ULN64               rc = deflt;
   TXA_OPTION         *o;

   ENTER();

   if ((o = TxaGetOption( txh, opt)) != NULL)
   {
      if (o->type == TXA_NUMBER)
      {
         rc = o->value.number;
      }
      else                                      // other option type
      {
         if (error != NULL)                     // give error/warning
         {
            TxPrint( "\n%s %s: %s'-%c'%s has no NUMBER value, using "
                     "default: 0x%llx = %llu\n",
                     (txh != TXA_1ST) ? "Option" : "Switch",
                      error, CBC, opt, CNN, deflt, deflt);
         }
      }
   }
   RETN64 (rc);
}                                               // end 'TxaOptionNum'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Get option value num/string, with bytes/kilo/mega/giga modifier and default
/*****************************************************************************/
ULN64 TxaOptionBkmg                             // RET   number value in bytes
(
   TXHANDLE            txh,                     // IN    TXA handle
   char                option,                  // IN    option character
   ULN64               def,                     // IN    default value
   BYTE                mod                      // IN    b,kb,mb,gb modifier
)
{
   ULN64               rc = 0;                  // function return
   BYTE                unit = TXA_DFUNIT;
   TXA_OPTION         *opt;                     // option pointer

   ENTER();
   TRACES(("Option: '%c' default: %llx mod:%2.2hx = '%c'\n",
            option, def, mod, mod));

   if ((opt = TxaOptValue( option)) != NULL)    // get the option details
   {
      switch (opt->type)
      {
         case TXA_STRING:
            rc   = TxaParseNumber( opt->value.string, TX_RADIX_STD_CLASS, &unit);
         case TXA_NO_VAL:
            rc   = def;
            break;

         default:                               // convert, default is MiB!
            rc   = opt->value.number;
            unit = opt->unit;
            break;
      }
   }
   if (unit == TXA_DFUNIT)
   {
      unit = mod;
   }
   switch (tolower(unit))
   {
      case 't': rc *= 1024;                     // Tera
      case 'g': rc *= 1024;                     // Giga
      case 'm': rc *= 1024;                     // Mega
      case 'k': rc *= 1024;                     // Kilo
      default:  break;                          // Bytes, OK
   }
   RETN64 (rc);
}                                               // end 'TxaOptionBkmg'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Get reference to specified argument-string
/*****************************************************************************/
char *TxaGetArgv                                // RET   argv string or ""
(
   TXHANDLE            txh,                     // IN    TXA handle
   int                 argn                     // IN    argument number 0..n
)
{
   char               *rc = "";                 // return empty if not there
   TXA_ELEMENT        *txa;                     // specified TXA instance

   if ((txa = txaHandle2Element( txh)) != NULL)
   {
      if ((argn >= 0) && (argn < txa->argc))    // valid argument nr ?
      {
         rc = txa->argv[argn];
         TRACES(("GETARGV argument: %d = '%s'\n", argn, rc));
      }
   }
   return( rc);
}                                               // end 'TxaGetArgv'
/*---------------------------------------------------------------------------*/

/*****************************************************************************/
// Get the number of arguments for specified TXA instance
/*****************************************************************************/
int TxaGetArgc                                  // RET   argument count
(
   TXHANDLE            txh                      // IN    TXA handle
)
{
   int                 rc = 0;
   TXA_ELEMENT        *txa;                     // specified TXA instance

   if ((txa = txaHandle2Element( txh)) != NULL)
   {
      rc = txa->argc;
   }
   return( rc);
}                                               // end 'TxaGetArgc'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Get the number of options for specified TXA instance
/*****************************************************************************/
int TxaGetOptc                                  // RET   option count
(
   TXHANDLE            txh                      // IN    TXA handle
)
{
   int                 rc = 0;
   TXA_ELEMENT        *txa;                     // specified TXA instance

   ENTER();
   if ((txa = txaHandle2Element( txh)) != NULL)
   {
      rc = txa->optc;
   }
   RETURN( rc);
}                                               // end 'TxaGetOptc'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Take handle value and convert to element, validate signature
/*****************************************************************************/
TXA_ELEMENT *txaHandle2Element                  // RET   element ptr or NULL
(
   TXHANDLE            txh                      // IN    TXA handle
)
{
   TXA_ELEMENT        *txa;

   if      (txh == TXA_CUR)   txa =                 txacur;
   else if (txh == TXA_PREV)  txa = (txacur) ?      txacur->prev : NULL;
   else if (txh == TXA_1ST)   txa =                 txa1st;
   else                       txa = (TXA_ELEMENT *) txh;

   if ((txa) && (txa->signature != TXA_SIGNATURE))
   {
      txa = NULL;
   }
   return( txa);
}                                               // end 'txaHandle2Element'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Copy item from string, honor embedded strings, strip quotes and escapes
/*****************************************************************************/
char *txaCopyItem                               // RET   copied string
(
   char               *destin,                  // OUT   destination buffer
   char               *start,                   // IN    start position
   int                 length                   // IN    max length to copy
)
{
   char               *s;
   char               *d;
   int                 l;
   char                delim  = ' ';            // default delimitter
   BOOL                escape = FALSE;

   if ((start != NULL) && (destin != NULL))
   {
      TRACES(( "COPY: Max length:%3.3u, start:}}}% -*.*s{{{\n",
                length, length, length, start));
      for (s = start, d = destin, l = 0; l < length; l++, s++)
      {
         if (*s == delim)
         {
            if (delim == ' ')                   // end of item reached
            {
               break;
            }
            else if (((*s == '"') || (*s == '\'')) && (escape))
            {
               *d++ = *s;                       // copy the character
            }
            else                                // end-of-string
            {
               delim = ' ';                     // to end of item ...
            }
         }
         else if ((!escape)  &&                 // prev was no escape
                  (delim == ' ')  &&            // and not in a string yet
                 ((*s == '"') || (*s == '\''))) // double or single quote
         {
            delim = *s;                         // to end of string
         }
         else if ( (*s     == '\\') &&          // possible escape
                  ((*(s+1) == ';')  ||          // for comment
                   (*(s+1) == '"')  ||          // for double quote
                   (*(s+1) == '\'') ))          // or single quote
         {
            escape = TRUE;                      // indicate escape next char
            continue;                           // and don't reset too early
         }
         else                                   // normal character
         {
            *d++ = *s;                          // copy the character
         }
         escape = FALSE;
//       TRACES(( "*s:'%c', esc:%u, destin:'%s'\n", *s, escape, destin));
      }
      *d = '\0';                                // terminate the string
      TRACES(("COPY : '%s'\n", destin));
   }
   return (destin);
}                                               // end 'txaCopyItem'
/*---------------------------------------------------------------------------*/

